;;;; -*- Mode: Lisp; indent-tabs-mode: nil -*-
;;;
;;; --- Wrapper over lfp_spawn(3)
;;;

#+xcvb (module (:depends-on ("ffi-types-unix")))

(in-package :iolib.os)

;;; Child process reaping protocol
;; NOTE: this protocol is NOT thread-safe.
;; If multiple threads use this protocol,
;; it is up to them to synchronize using a lock.

(defvar *child-process-reaper-hook*
  ())

(defun register-child-reaper (reaper)
  "Register REAPER as a function to handle dead children processes.
REAPER must be a designator for a function taking two arguments and
returning a generalized boolean. The arguments are a PID and a STATUS
to be decoded with such functions as ISYS:WIFEXITED and ISYS:WEXITSTATUS.
The return value is true iff the reaper could indeed handle specified process."
  (pushnew reaper *child-process-reaper-hook*))

(defun unregister-child-reaper (reaper)
  (setf *child-process-reaper-hook*
        (delete reaper *child-process-reaper-hook*)))

(defun reap-child (pid status)
  "Call each registered reaper until one is found that matches the child process."
  (let ((hook *child-process-reaper-hook*))
    (if (null hook)
        (format *error-output*
                "No registered reaper to take care of dead child, pid: ~D status: #x~X"
                pid status)
        (let ((handled
               (loop :for reaper :in hook :thereis (funcall reaper pid status))))
          (unless (or handled (isys:wifstopped status) (isys:wifcontinued status))
            (format *error-output*
                    "None of the registered reapers could handle dead child, pid: ~D status: #x~X"
                    pid status))
          handled))))

(defun reap-one-child (&key options nohang)
  "Reaps a child, if any is ready, using the reap-child protocol.
waitpid is called with options, and additionally WNOHANG if nohang is true.
Returns two values: reaped-p and more-children-p.
reaped-p is T is a child was reaped, NIL if none was reaped.
more-children-p is
T if there are more children (because you used nohang, but no child is ready),
:UNKNOWN if it's unknown whether there are more (because we reaped one),
NIL if there are no more children to wait for (and you had no business reaping).
"
  (block nil
    (let* ((options (or options 0))
           (options (if nohang (logior options isys:wnohang) options)))
      (multiple-value-bind (pid status)
          (handler-case
              (isys:waitpid -1 options)
            (isys:echild () ;; No more children
              (return (values nil nil))))
        (when (zerop pid) ;; More active children
          (return (values nil t)))
        (reap-child pid status)
        (values t :unknown)))))

(defun reap-children (&key options nohang max-children)
  "For each child the status of which is available through waitpid
   (with given waitpid OPTIONS), invoke HANDLER with pid and status
   (to be decoded with isys::wifexited and friends).
   Return two values, more-children-p and count.
   The first value is T if there are still active children (but none was ready and we don't hang),
   NIL if there are no more, :UNKNOWN if it's not known (because we reached the max count).
   The second value is a count of how many children were processed.
   NOTE1: this function MUST be run inside a WITH-WAITPID-LOCK.
   NOTE2: if this has any chance of running in a signal handler,
   then both WITH-WAITPID-LOCK and WITH-CHILD-PROCESS-LOCK need to
   be protected by some signal-blocking unwind-protect."
  (loop :for count :from 0 :do
    (when (and max-children (>= count max-children))
      (return (values :unknown count)))
    (multiple-value-bind (reapedp more-children-p)
        (reap-one-child :options options :nohang nohang)
      (unless reapedp
        (return (values more-children-p count))))))


;;; Process

(defun tty-read-fn (fd buf nbytes)
  (handler-case
      (isys:read fd buf nbytes)
    (isys:eio () 0)))

(defun tty-write-fn (fd buf nbytes)
  (handler-case
      (isys:write fd buf nbytes)
    (isys:eio ()
      (error 'isys:epipe
             :handle fd
             :syscall "write"))))

(defclass tty-stream (iolib.streams:dual-channel-gray-stream)
  ()
  (:default-initargs :read-fn  #'tty-read-fn
                     :write-fn #'tty-write-fn))

(defclass process ()
  ((pid    :initarg :pid :reader process-pid)
   (status :initform :running) ; exit status. nil if not reaped yet.
   (closed :initform nil :reader process-closed) ; have we already closed the process and its standard fds?
   (stdin  :reader process-stdin)
   (stdout :reader process-stdout)
   (stderr :reader process-stderr)
   (pty    :reader process-pty)
   (fdmap  :initarg :fdmap :reader process-fdmap)))

(defun process-stdin-fd (process)
  (values (gethash +stdin+ (process-fdmap process))))
(defun process-stdout-fd (process)
  (values (gethash +stdout+ (process-fdmap process))))
(defun process-stderr-fd (process)
  (values (gethash +stderr+ (process-fdmap process))))

(defmethod initialize-instance :after ((process process) &key
                                       external-format ptmfd)
  (flet ((make-stream (fd)
           (make-instance 'tty-stream :fd fd
                          :external-format external-format)))
    (macrolet ((init (slot fd)
                 `(with-slots (,slot) process
                    (when ,fd (setf ,slot (make-stream ,fd)))))
               (init-std (slot)
                 (let ((fd-accessor
                        (intern (format nil "~A~A~A" '#:process- slot '#:-fd))))
                   `(let ((fd (,fd-accessor process)))
                      (init ,slot fd)))))
      (init-std stdin) (init-std stdout) (init-std stderr) (init pty ptmfd))))

(defmethod close ((process process) &key abort)
  (if (process-closed process)
      nil
      (macrolet ((close-process-stream (slot)
                   `(when (slot-boundp process ',slot)
                      (close (slot-value process ',slot) :abort abort)
                      (slot-makunbound process ',slot))))
        ;; Note: making the slot unbound rather than nil, to avoid
        ;; (read-line (process-stdout p)) doing the wrong thing.
        (close-process-stream stdin)
        (close-process-stream stdout)
        (close-process-stream stderr)
        (close-process-stream pty)
        (setf (slot-value process 'closed) t)
        (when (not abort)
          (process-wait process))
        t)))

(defmethod print-object ((o process) s)
  (print-unreadable-object (o s :type t :identity nil)
    (format s "~S ~S ~S ~S"
            :pid (process-pid o)
            :status (process-status o))))

(defun exit-status (status)
  (cond
    ((isys:wifexited status)
     (isys:wexitstatus status))
    ((isys:wifsignaled status)
     (values (isys:wtermsig* status)
             (isys:wcoredump status)))
    (t
     (error "improper exit status ~S" status))))

(defgeneric process-status (process &key wait)
  (:method ((process process) &key wait)
    (when wait
      (process-wait process))
    (with-slots (status) process
      (if (integerp status)
          (exit-status status)
          :running))))

(defgeneric process-activep (process)
  (:method ((process process))
    (eql :running (process-status process))))

(defgeneric process-kill (process &optional signum)
  (:method ((process process) &optional (signum :sigterm))
    (isys:kill (process-pid process) signum)
    process))


(defmacro with-lfp-spawn-arguments ((attributes file-actions pid) &body body)
  (with-gensyms (spawnattr-initialized-p file-actions-initialized-p)
    `(with-foreign-objects ((,attributes 'lfp-spawnattr-t)
                            (,file-actions 'lfp-spawn-file-actions-t)
                            (,pid 'pid-t))
       (let ((,spawnattr-initialized-p nil)
             (,file-actions-initialized-p nil))
         (unwind-protect
              (progn
                (setf ,spawnattr-initialized-p
                      (lfp-spawnattr-init ,attributes))
                (setf ,file-actions-initialized-p
                      (lfp-spawn-file-actions-init ,file-actions))
                ,@body)
           (when ,spawnattr-initialized-p
             (lfp-spawnattr-destroy ,attributes))
           (when ,file-actions-initialized-p
             (lfp-spawn-file-actions-destroy ,file-actions)))))))

(defun allocate-argv (argv program arglist)
  ;; copy program name
  (setf (mem-aref argv :pointer 0)
        (foreign-string-alloc program))
  ;; copy program arguments
  (loop :for i :from 1
        :for arg :in arglist :do
        (setf (mem-aref argv :pointer i)
              (foreign-string-alloc arg))))

(defun find-program (program)
  (cond
    ((eql :shell program)
     (list "/bin/sh" "-c"))
    (t
     (file-path-namestring program))))

(defmacro with-argv (((arg0 argv) program arguments) &body body)
  (with-gensyms (argc)
    `(let* ((,program (ensure-list (find-program ,program)))
            (arguments (append (cdr ,program) arguments))
            (,argc (+ 2 (length ,arguments))))
       (with-foreign-object (,argv :pointer ,argc)
         (isys:bzero ,argv (* ,argc (isys:sizeof :pointer)))
         (unwind-protect
              (progn
                (allocate-argv ,argv (car ,program) ,arguments)
                (let ((,arg0 (mem-ref ,argv :pointer)))
                  ,@body))
           (delocate-null-ended-list ,argv))))))

(defun effect-redirections (results parent-close
                            file-actions spec &key (close-other-fds t) external-format ptmfd pts)
  "RESULTS is an equal hash-table the keys of which are file handles on the child side,
the values of which are the corresponding file handles on the parent side.
PARENT-CLOSE is an equal hash-table the keys of which are file handles to close on the parent.
FILE-ACTIONS is an object with which to call lfp-spawn-file-action-*
SPEC is a list of (FD . REDIR) commands.
Returns a hash-table mapping subprocess FD to results."
  (declare (ignore external-format)) ;; FOR NOW - TODO: always open streams, with external-format
  (let ((shuffle (make-hash-table :test 'equal)) ; mapping from child fd to parent fd
        (unshuffle (make-hash-table :test 'equal)) ; mapping from parent fd to child fd
        (dups (make-hash-table :test 'equal)) ; duplicates after redirections
        (open-max (isys:open-max))) ; total number of fd's
    (loop :for (fd . redir) :in spec :do
      (check-type fd unsigned-byte)
      (assert (< fd open-max))
      (check-type redir (or symbol unsigned-byte cons))
      (when (integerp redir) (assert (< redir open-max)))
      (effect-redirection results shuffle unshuffle dups parent-close fd redir ptmfd pts))
    (loop :for fd :being :the :hash-keys :of dups :using (:hash-value dup) :do
      (assert (gethash dup shuffle)))
    ;; close all fd's that haven't been explicitly redirected
    ;; TODO: use introspection via /proc/self/fd to minimize number of closes?
    (when close-other-fds
      (loop :for fd :below open-max
        :when (fd-open-p fd)
        :unless (nth-value 1 (gethash fd shuffle)) :do
        (setf (gethash fd shuffle) nil)))
    (loop
      :with child-fds = (loop :for fd :being :the :hash-keys :of shuffle :collect fd)
      :for fd :in child-fds
      :when (nth-value 1 (gethash fd shuffle)) :do
      (let ((basefd
             (loop
               ;; The linearity of redirections, enforced in effect-redirection,
               ;; ensures that any circularity will be an actual circle and goes
               ;; back to the starting point. No loops that start at one ancestor of it,
               ;; no multiple loops that share multiple nodes.
               :for prev = fd :then xfd
               :for xfd = (gethash prev unshuffle) :do
               (cond
                 ((null xfd)
                  (return prev))
                 ((= fd xfd)
                  (return nil))))))
        (cond
          ((equal fd (gethash fd shuffle))
           (remhash fd shuffle))
          (t
           (when (null basefd) ; break circularity by duplicating a fd
             (let ((tempfd (isys:dup fd))
                   (xfd (gethash fd unshuffle)))
               (setf (gethash xfd shuffle) tempfd)
               (setf (gethash tempfd unshuffle) xfd)
               (setf (gethash tempfd parent-close) nil)
               (setf basefd fd)))
           (loop :for targetfd = basefd :then source
             :for source = (prog1 (gethash targetfd shuffle) (remhash targetfd shuffle)) :do
             (etypecase source
               (unsigned-byte
                (unless (= source targetfd)
                  (lfp-spawn-file-actions-adddup2 file-actions source targetfd)))
               (null
                (lfp-spawn-file-actions-addclose file-actions targetfd)
                (return))
               (cons
                (destructuring-bind (path flags mode) source
                  (check-type path string)
                  (check-type flags integer)
                  (check-type mode integer)
                  (lfp-spawn-file-actions-addopen file-actions targetfd path flags mode))
                (return))))))))
    (loop :for targetfd :being :the :hash-keys :of dups :using (:hash-value source) :do
      (lfp-spawn-file-actions-adddup2 file-actions source targetfd))))

(defun normalize-open-flags (oflags &optional mode)
  (etypecase oflags
    (unsigned-byte
     oflags)
    ((member nil :input)
     isys:o-rdonly)
    ((eql :output)
     (logior isys:o-wronly (if mode isys:o-creat 0)))
    ((eql :append)
     (logior isys:o-wronly isys:o-append (if mode isys:o-creat 0)))
    ((eql :io)
     (logior isys:o-rdwr (if mode isys:o-creat 0)))))

(defun normalize-open-mode (mode)
  (etypecase mode
    (unsigned-byte
     mode)
    (null
     0)
    (t
     #o666)))

(defun close-fds (&rest fds)
  (dolist (fd fds)
    (when fd (isys:close fd))))

(defun setup-slave-pty (new-ctty-p)
  (if new-ctty-p
      (let ((ptmfd (isys:openpt (logior isys:o-rdwr isys:o-noctty isys:o-cloexec))))
        (isys:grantpt ptmfd)
        (isys:unlockpt ptmfd)
        (values ptmfd (isys:ptsname ptmfd)))
      (values nil nil)))

(defmacro with-pty ((new-ctty-p ptmfd pts) &body body)
  `(multiple-value-bind (,ptmfd ,pts)
       (setup-slave-pty ,new-ctty-p)
     (unwind-protect
          (locally ,@body)
       (unless ,new-ctty-p
         (close-fds ,ptmfd)))))

(defun process-other-spawn-args (attributes new-session pts current-directory
                                 uid gid resetids)
  (when new-session
    (lfp-spawnattr-setsid attributes))
  (when pts
    (lfp-spawnattr-setctty attributes pts))
  (when current-directory
    (lfp-spawnattr-setcwd attributes current-directory))
  (when uid
    (lfp-spawnattr-setuid attributes uid))
  (when gid
    (lfp-spawnattr-setgid attributes gid))
  (when resetids
    (lfp-spawnattr-setflags attributes lfp-spawn-resetids)))

(defun ensure-hash (key table value)
  (assert (not (nth-value 1 (gethash key table))))
  (setf (gethash key table) value))

(defun effect-open-redirection (shuffle unshuffle parent-close
                                fd path flags mode)
  (declare (ignorable unshuffle parent-close))
  ;; #+trust-addopen (ensure-hash fd shuffle (list path flags mode))
  ;; #-trust-addopen
  (let ((redir (isys:open path flags mode)))
    (ensure-hash fd shuffle redir)
    (unless (= fd redir)
      (ensure-hash redir shuffle nil))
    (ensure-hash redir unshuffle fd)
    (ensure-hash redir parent-close nil)))

(defun effect-redirection (results shuffle unshuffle dups parent-close fd redir ptmfd pts)
  (etypecase redir
    (unsigned-byte
     (ensure-hash fd shuffle redir)
     (ensure-hash redir unshuffle fd))
    (iolib.streams:dual-channel-fd-mixin
     (let ((rfd (iolib.streams:fd-of redir)))
       (ensure-hash fd shuffle rfd)
       (ensure-hash rfd unshuffle fd)))
    (null
     (ensure-hash fd shuffle nil))
    ((eql :stdout)
     (ensure-hash fd shuffle nil)
     (ensure-hash fd dups 1))
    ((member :null :null-io)
     (effect-open-redirection
      shuffle unshuffle parent-close
      fd "/dev/null" isys:o-rdwr 0))
    ((eql :null-in)
     (effect-open-redirection
      shuffle unshuffle parent-close
      fd "/dev/null" isys:o-rdonly 0))
    ((eql :null-out)
     (effect-open-redirection
      shuffle unshuffle parent-close
      fd "/dev/null" isys:o-wronly 0))
    ((eql :pty-in)
     (ensure-hash ptmfd results fd)
     (effect-open-redirection
      shuffle unshuffle parent-close
      fd pts isys:o-rdonly 0))
    ((eql :pty-out)
     (ensure-hash ptmfd results fd)
     (effect-open-redirection
      shuffle unshuffle parent-close
      fd pts isys:o-wronly 0))
    ((eql :pty)
     (ensure-hash ptmfd results fd)
     (effect-open-redirection
      shuffle unshuffle parent-close
      fd pts isys:o-rdwr 0))
    (cons ; path open-flag open-mode
     (destructuring-bind (path &optional oflags mode) redir
       (effect-open-redirection
        shuffle unshuffle parent-close
        fd (file-path-namestring path)
        (normalize-open-flags oflags mode)
        (normalize-open-mode mode))))
    ((eql t)
     (ensure-hash fd shuffle fd)
     (ensure-hash fd unshuffle fd))
    ((eql :pipe-in)
     (multiple-value-bind (read-fd write-fd) (isys:pipe)
       (ensure-hash fd shuffle read-fd)
       (ensure-hash read-fd unshuffle fd)
       (ensure-hash fd results write-fd)
       (ensure-hash read-fd parent-close nil)))
    ((eql :pipe-out)
     (multiple-value-bind (read-fd write-fd) (isys:pipe)
       (ensure-hash fd shuffle write-fd)
       (ensure-hash write-fd unshuffle fd)
       (ensure-hash fd results read-fd)
       (ensure-hash write-fd parent-close nil)))))

(defun effect-signal-attributes (attributes sigmask sigdefault)
  (when sigmask
    (with-foreign-object (%set 'isys:sigset-t)
      (isys:encode-sigset %set sigmask)
      (lfp-spawnattr-setsigmask attributes %set)))
  (when sigdefault
    (with-foreign-object (%set 'isys:sigset-t)
      (isys:encode-sigset %set sigdefault)
      (lfp-spawnattr-setsigdefault attributes %set)))
  (values))

(defmacro with-redirections ((results to-be-closed) &body body)
  `(let ((,results (make-hash-table :test 'equal))
         (,to-be-closed (make-hash-table :test 'equal)))
     (unwind-protect-case ()
         (locally ,@body)
       (:always
        (apply 'close-fds (hash-table-keys ,to-be-closed)))
       (:abort
        (apply 'close-fds (hash-table-values ,results))))))

(defvar *children-processes* (make-hash-table :test 'equal))

(defun create-process-child-reaper (pid status)
  (unless (or (isys:wifstopped status) (isys:wifcontinued status))
    (let ((process (gethash pid *children-processes*)))
      (when process
        (remhash pid *children-processes*)
        (with-slots ((xpid pid) (xstatus status)) process
          (setf xpid nil)
          (setf xstatus status))
        t))))

;; program-and-args: a list of strings, or a single string meaning a singleton list.
;;    The first string is the path to the program; it can be :shell for the system shell
;;    (or something to search in $PATH if relative),
;;    the rest is the arguments.
;; argv0: a replacement for the program path to be passed in argv[0]
;; environment: t - inherit environment
;;              nil - NULL environment
;;              alist - the environment to use
;; redirect: an alist mapping fd number in the child to redirection
;;   each redirection may be
;;         file-path designator - open file, redirect to it
;;         file-descriptor designator(integer or stream) - file descriptor in parent, redirect to it
;;         :pipe-in :pipe-out - create pipe, redirect the child descriptor to one end
;;                and wrap the other end into a stream the fd of which goes into the process-fdmap.
;;         :null :null-io :null-in :null-out - redirect to bit bucket, rdwr, rdonly, wronly.
;;                useful because /dev/null doesn't exist on Windows
;;         :pty - open a pty, master fd goes to the process-fdmap, the child gets the slave fd.
;;         t - inherit
;;         nil - close
;; stdin, stdout, stderr:
;;         specify a redirection as if redirecting fd 0, 1 and 2 respectively above.
;;         additionally, the following redirection is recognized:
;;         :pipe, :null, file-path designator: will redirect in the standard direction,
;;         .i.e. input for stdin, output for stdout and stderr.
;;         (creating file if output direction and it doesn't exist)
;; pty: boolean - spawn a new controlling tty. it is also implicitly T if
;;                either stdin, stdout or stderr is :pty
;; new-session: boolean - create a new session using setsid(). it is also implicitly T
;;                        if a PTY is requested
;; current-directory: path - a directory to switch to before executing
;; uid: user id - unsigned-byte or string
;; gid: group id - unsigned-byte or string
;; resetids: boolean - reset effective UID and GID to saved IDs
;; sigmask, sigdefault: change the signal settings in the spawned process

(defun create-process (program-and-args &key (environment t)
                       redirect (stdin :pipe) (stdout :pipe) (stderr :pipe) pty
                       sigmask sigdefault
                       new-session current-directory uid gid resetids
                       (external-format :utf-8))
  ;; stdin, stdout, stderr: a redirection, with additional autodetection
  ;;         :stdout - for stderr, indicates it must copy stdout.
  (register-child-reaper 'create-process-child-reaper)
  (destructuring-bind (program &rest arguments)
      (ensure-list program-and-args)
    (with-argv ((arg0 argv) program arguments)
      (with-c-environment (envp environment)
        (with-lfp-spawn-arguments (attributes file-actions pid)
          (let ((new-ctty-p (or pty (member :pty redirect :key 'second))))
            (with-pty (new-ctty-p ptmfd pts)
              (flet ((normalize-input-redirection (x)
                       (etypecase x
                         ((eql :pipe) :pipe-in)
                         ((eql :null) :null-in)
                         ((eql :pty) :pty) ;; pty-in ?
                         ((or cons unsigned-byte symbol) x)
                         ((or string pathname) (list x :input))))
                     (normalize-output-redirection (x)
                       (etypecase x
                         ((eql :pipe) :pipe-out)
                         ((eql :null) :null-out)
                         ((eql :pty) :pty) ;; pty-out ?
                         ((or cons unsigned-byte symbol) x)
                         ((or string pathname) (list x :output t)))))
                (with-redirections (results to-be-closed)
                  (effect-redirections
                   results to-be-closed
                   file-actions
                   (append
                    (when stdin (list (cons +stdin+ (normalize-input-redirection stdin))))
                    (when stdout (list (cons +stdout+ (normalize-output-redirection stdout))))
                    (when stderr (list (cons +stderr+ (normalize-output-redirection stderr))))
                    redirect)
                   :ptmfd ptmfd :pts pts :external-format external-format)
                  (effect-signal-attributes attributes sigmask sigdefault)
                  (process-other-spawn-args
                   attributes (or new-session new-ctty-p) pts
                   current-directory uid gid resetids)
                  (lfp-spawnp pid arg0 argv envp file-actions attributes)
                  (let* ((pid (mem-ref pid 'pid-t))
                         (process (make-instance 'process
                                                 :pid pid :fdmap results :ptmfd ptmfd
                                                 :external-format external-format)))
                    (setf (gethash pid *children-processes*) process)
                    process))))))))))

(defun slurp-char-stream (stream)
  (with-output-to-string (s)
    (loop :for c := (read-char stream nil nil)
          :while c :do (write-char c s))))

(defun slurp-octet-stream (stream)
  (let ((dynbuffer (make-instance 'dynamic-buffer :size 4096 :growth-size 2))
        (iobuffer (make-array 4096 :element-type '(unsigned-byte 8))))
    (handler-case
        (loop :for pos := (read-sequence iobuffer stream)
              :while (plusp pos) :do
          (write-vector dynbuffer iobuffer 0 pos))
      (end-of-file () nil))
    (subseq (sequence-of dynbuffer)
            (read-cursor-of dynbuffer)
            (write-cursor-of dynbuffer))))

(defun run-program (program-and-args &key (environment t)
                    (stdin :null) (stdout :pipe) (stderr :pipe)
                    (external-format :utf-8))
  (flet ((slurp (stream)
           (if external-format
               (slurp-char-stream stream)
               (slurp-octet-stream stream)))
         (make-empty-output ()
           (if external-format
               (make-array 0 :element-type 'character)
               (make-array 0 :element-type '(unsigned-byte 8)))))
    (let ((process
            (create-process program-and-args
                            :environment environment
                            :stdin stdin
                            :stdout stdout
                            :stderr stderr
                            :external-format external-format)))
      (unwind-protect
           (let ((stdout (if (eql :pipe stdout)
                             (slurp (process-stdout process))
                             (make-empty-output)))
                 (stderr (if (eql :pipe stderr)
                             (slurp (process-stderr process))
                             (make-empty-output))))
             (values (process-status process :wait t)
                     stdout stderr))
        (close process)))))

(defmethod process-wait ((process process))
  (loop :while (process-activep process) :do
    (reap-children))
  (process-status process))

(defmethod process-kill ((process process) &optional (signum :sigterm))
  (isys:kill (process-pid process) signum))
